#include "parti.h"
#include "vararg.h"

/* variable-length arguments 
 *
 * new_mulgather(sched, size, narray, data, target, ...)
 * SCHED *sched;		schedule
 * int   size;			size of each element
 * int   narray;		number of arrays
 * char  *data,*target;		arrays of data arrays
 */
void 
PARTI_mulgather VA_DCL (void *, args)
{
  int   i, j;
  int   size;
  int   narray;
  char  **data, **target;
  SCHED *sched;
  va_list ap;
  void  new_multi_gather();

  /* retrieve the parameters from the list */
  sched  = (SCHED *) VA_START (ap, args, void *);
  size   = (int) va_arg (ap, int);
  narray = (int) va_arg (ap, int);
  data   = (char **) malloc (sizeof(char *)*narray);
  target = (char **) malloc (sizeof(char *)*narray);
  for (i = 0; i < narray; i++) {
    data[i]   = (char *) va_arg (ap, char *);
    target[i] = (char *) va_arg (ap, char *);
  }

  /* call the multi_gather function */
  new_multi_gather(sched, size, narray, data, target);

  /* free the allocated memory */
  free (target);
  free (data);
}

/*
 *  wrapper for the mulgather function
 */

extern SCHED *get_sched_ptr();

void 
parti_mulgather_ VA_DCL (void *, args)
{
  int   i, j;
  int   *size;
  int   *narray;
  int   *sched;
  char  **data, **target;
  va_list ap;
  void  new_multi_gather();

  /* retrieve the parameters from the list */
  sched  = (int *) VA_START (ap, args, int *);
  size   = (int *) va_arg (ap, int *);
  narray = (int *) va_arg (ap, int *);
  data   = (char **) malloc (sizeof(char *)*(*narray));
  target = (char **) malloc (sizeof(char *)*(*narray));
  for (i = 0; i < *narray; i++) {
    data[i]   = (char *) va_arg (ap, char *);
    target[i] = (char *) va_arg (ap, char *);
  }

  /* call the multi_gather function */
  new_multi_gather((SCHED *)get_sched_ptr(*sched),*size,*narray,data,target);

  /* free the allocated memory */
  free (target);
  free (data);
}

/*
 * Multigather function
 */
void
new_multi_gather(sched, size, narray, data, target)
SCHED *sched;			/* schedule */
int   size;			/* size of each element */
int   narray;			/* number of arrays */
char  **data,**target;		/* arrays of data arrays */
{
  int  count,i,j,k,n,m;
  char **blist,*buffer;
  char *ptr;
  int  *ibuf, *idata;
  double *dbuf, *ddata;

#ifdef DEBUG
  trace_enter("new_mulgather");
#endif
  blist = wbuffer->blist;
  buffer = (char *) malloc(size*narray*sched->ndata);
  count = 0;
  for(i=0;i<PARTI_numnodes;i++){
      blist[i] = &buffer[count];
      count += size * sched->sr->fetch_n[i] * narray;
  }

  new_multi_execute_gather(sched->sr, size, narray, blist, target);

  switch (size) {
  case sizeof(int):
    ibuf = (int *) buffer;
    for (i = 0, count = 0, m = 0; i < PARTI_numnodes; i++) {
      for (j = 0; j < narray; j++) {
	idata = (int *) data[j];
        for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
          idata[sched->perm[n]] = ibuf[m];
      }
      count += sched->sr->fetch_n[i];
    }
  case sizeof(double):
    dbuf = (double *) buffer;
    for (i = 0, count = 0, m = 0; i < PARTI_numnodes; i++) {
      for (j = 0; j < narray; j++) {
	ddata = (double *) data[j];
        for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
          ddata[sched->perm[n]] = dbuf[m];
      }
      count += sched->sr->fetch_n[i];
    }
  case sizeof(char):
    for (i = 0, count = 0, m = 0; i < PARTI_numnodes; i++) {
      for (j = 0; j < narray; j++) {
        for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
          data[j][sched->perm[n]] = buffer[m];
      }
      count += sched->sr->fetch_n[i];
    }
  default:
    for (i = 0, ptr = buffer, count = 0; i < PARTI_numnodes; i++) {
      for (j = 0; j < narray; j++) {
        for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, ptr+=size)
          bytecopy(&data[j][size*sched->perm[n]], ptr, size);
      }
      count += sched->sr->fetch_n[i];
    }
  }

  free(buffer);

#ifdef DEBUG
  trace_exit();
#endif
}
